
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* malloc() free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>

#include "initng_global.h"
#include "initng_handler.h"
#include "initng_history.h"
#include "process_db.h"
#include "colorprint.h"
#include "initng_execute.h"

#ifdef STATIC_I_PARSER
#include "initng_i_parser.h"
#endif
#ifdef XML_PARSER
#include "initng_xml_parser.h"
#endif
#ifdef STATIC_RL_PARSER
#include "initng_rl_parser.h"
#endif

#include "main.h"
#include "service_db.h"
#include "active_db.h"
#include "initng_load_module.h"

/* function defines */
static int load_all_deps(active_h * process_starting);
static int service_blacklisted(char *name);
static void handle_killed_start(active_h * killed_process);
static void handle_killed_stop(active_h * killed_service);
static int try_to_start_service(active_h * service_to_start);
static int try_to_stop_service(active_h * service_to_stop);
static void mark_service(active_h * service, e_a_status status);
static int is_use_up(active_h * service_starting);
static int get_service(active_h * service);
static void system_up(void);

/* this function trys to find the from_service pointer from a service name.
   this actually loads the config from disk, parser by parser */
static int get_service(active_h * service)
{
    assert(service);
    assert(service->name);
    D_("get_service(%s);\n", service->name);

    /* check if service is set already */
    if (service->from_service)
    {
        D_("get_service(%s): already have a service!\n", service->name);
        return (TRUE);
    }

    /* check cache database */
    if ((service->from_service = service_db_find_by_name(service->name)))
    {
        D_("get_service(%s): found service in service database!\n",
           service->name);
        return (TRUE);
    }

    /* try parse with dynamic loaded parsers */
    if ((service->from_service = initng_modules_parse(service->name)))
    {
        D_(" get_service(%s): got service from initng dynamic parsers.\n",
           service->name);
        return (TRUE);
    }

#ifdef STATIC_XML_PARSER
    /* get initng xml loader a chance */
    if ((service->from_service = initng_xml_load_process(service->name)))
    {
        D_("get_service(%s): got service from initng xml loader.\n",
           service->name);
        return (TRUE);
    }
#endif

#ifdef STATIC_I_PARSER
    /* get initng loader a chance */
    if ((service->from_service = initng_i_parser(service->name)))
    {
        D_("get_service(%s): got service from initng loader.\n",
           service->name);
        return (TRUE);
    }
#endif

#ifdef STATIC_RL_PARSER
    /* get runlevel loader a chance */
    if ((service->from_service = initng_rl_parser(service->name)))
    {
        D_("get_service(%s): got service from runlevel loader.\n",
           service->name);
        return (TRUE);
    }
#endif

    /* or fail */
    D_("get_service(%s): Can't get source.. \n", service->name);
    return (FALSE);
}

/* creates a service */
int load_to_active(active_h ** a_new, char *service_name)
{
    assert(service_name);
    D_("load_to_active(%s); \n", service_name);

    /* check the blacklist db */
    if (service_blacklisted(service_name))
    {
        F_("load_to_active(%s): Service BLACKLISTED.\n", service_name);
        return (FALSE);
    }

    /* Make sure there isn't any process with this name running */
    if ((*a_new = active_db_find_by_name(service_name)))
    {
        D_("load_to_active(%s): process exists, returning pointer to it. \n",
           service_name);
        return (TRUE);
    }

    /* if not create a new active entry, out of memory? */
    if (!active_db_new(a_new, service_name))
    {
        D_("load_to_active(%s): Unable to allocate process, out of memory?\n",
           service_name);
        (*a_new) = NULL;
        return (FALSE);
    }

    /* this actually loads service from disk */
    if (!get_service(*a_new) || !(*a_new)->from_service)
    {
        F_("load_to_active(%s): Can't set service!\n", service_name);
        active_db_free(*a_new);
        (*a_new) = NULL;
        return (FALSE);
    }

    if ((*a_new)->from_service->type != TYPE_SERVICE &&
        (*a_new)->from_service->type != TYPE_RUNLEVEL)
    {
        F_("load_to_active(%s): service is not a service type!\n",
           service_name);
        active_db_free(*a_new);
        (*a_new) = NULL;
        return (FALSE);
    }

#ifdef VERBOSE
#ifdef PRINT_SERVICE
    /* test output */
    print_service(s_new);
#endif
#endif

    /* if there are dependencies load them too */
    if (active_db_is(STRING_NEED, (*a_new)))
    {
        if (!load_all_deps(*a_new))
        {
            F_("load_to_active(%s): dependcy refuse to load!\n",
               service_name);
            active_db_free(*a_new);
            (*a_new) = NULL;
            return (FALSE);
        }
    }

    /* return with pointer to service */
    if (active_db_add(a_new))
        return (TRUE);
    else
        return (FALSE);

}

/* When this function is called, the service is marked for starting */
e_a_status start_service(active_h * service_to_start)
{
    assert(service_to_start);
    assert(service_to_start->name);

    /* STEP 2, service must be in these statuses to start */
    if (service_to_start->a_status != NONE &&
        service_to_start->a_status != STOPPED &&
        service_to_start->a_status != RESPAWNING)
        return (FAIL_STARTING);

    /* STEP 3, mark service starting */
    mark_service(service_to_start, LOADING);

    /* STEP 5, try mark-start all dependencies */
    if (service_to_start->depends)
    {
        int i;

        /* walk through and launch all */
        for (i = 0; service_to_start->depends[i]; i++)
            start_service(service_to_start->depends[i]);
        /* settle back, and wait for deps */
        mark_service(service_to_start, WAITING_FOR_START_DEP);
        return (WAITING_FOR_START_DEP);
    }

    /* no deps = all deps met */
    mark_service(service_to_start, START_DEP_MET);
    return (START_DEP_MET);
}

/* This is the function that really launch the start-marked services if all deps are up */
static int try_to_start_service(active_h * service_to_start)
{
    assert(service_to_start);

    D_("try_to_start_service(%s): status: \"%s\"\n",
       service_to_start->name,
       active_db_get_status_string(service_to_start->a_status));

    /* check if all deps is up = DONE or RUNNING */
    if (service_to_start->depends)
    {
        int i;

        for (i = 0; service_to_start->depends[i]; i++)
            if ((service_to_start->depends[i]->a_status != DONE) &&
                (service_to_start->depends[i]->a_status != RUNNING))
            {
                D_(" \"%s\" wait for --> \"%s\"(s:%i).\n",
                   service_to_start->name,
                   service_to_start->depends[i]->name,
                   service_to_start->depends[i]->a_status);
                return (FALSE);
            }
    }

    /* check if is has a start process */
    if (service_to_start->start_process)
    {
        F_("try_to_start_service(%s): Function already have a start process! Aborting\n", service_to_start->name);
        return (FALSE);
    }

    /* load a service if missing */
    if (!get_service(service_to_start))
        return (FALSE);

    /* do some checkings to the STRING_USE */
    if (!is_use_up(service_to_start))
    {
        D_("Use deps not up\n");
        return (FALSE);
    }

    /* F I N A L L Y   S T A R T */

    /* WE ARE EXECUTING A START FILE */
    if (active_db_is(STRING_START, service_to_start))
    {
        D_("try_to_start_service(%s): Found START....\n",
           service_to_start->name);
        if (!start_execute(service_to_start))
        {
            mark_service(service_to_start, FAIL_STARTING);
            return (FALSE);
        }
        /* should we set status to running directly? */
        if (active_db_is(ENABLED_UP_WHEN_STARTED, service_to_start))
            mark_service(service_to_start, RUNNING);
        else if (active_db_is(ENABLED_UP_WHEN_PID_SET, service_to_start))
            mark_service(service_to_start, STARTING);
        else
            mark_service(service_to_start, STARTING);
        return (TRUE);                      /* return with status */
    }

    /* WE ARE EXECUTING A START SCRIPT */
    if (active_db_is(STRING_START_SCRIPT, service_to_start))
    {
        D_("try_to_start_service(%s): Found START_SCRIPT....\n",
           service_to_start->name);
        if (!start_script_execute(service_to_start))
        {
            mark_service(service_to_start, FAIL_STARTING);
            return (FALSE);
        }
        /* should we set status to running directly? */
        if (active_db_is(ENABLED_UP_WHEN_STARTED, service_to_start))
            mark_service(service_to_start, RUNNING);
        else if (active_db_is(ENABLED_UP_WHEN_PID_SET, service_to_start))
            mark_service(service_to_start, STARTING);
        else
            mark_service(service_to_start, STARTING);

        return (TRUE);                      /* return with status */
    }

    /* we are executing a DAEMON */
    if (active_db_is(STRING_DAEMON, service_to_start))
    {
        D_("try_to_start_service(%s): Found DAEMON....\n",
           service_to_start->name);
        if (!daemon_execute(service_to_start))
        {
            mark_service(service_to_start, FAIL_STARTING);
            return (FALSE);
        }
        /* check if we have som data of when this
         * service will be declared up */
        if (active_db_is(ENABLED_UP_WHEN_DONE, service_to_start) ||
            active_db_is(ENABLED_UP_WHEN_PID_SET, service_to_start))
            mark_service(service_to_start, STARTING);
        else
            mark_service(service_to_start, RUNNING);
        return (TRUE);                      /* return with status */
    }

    /* D O N E - D I R E C T */
    mark_service(service_to_start, DONE);
    return (TRUE);                          /* return with status */
}

/* will try start processes waiting for dependencies to go up */
void try_start_wait_deps(void)
{
    active_h *tmp = NULL;

    D_("try_start_wait_deps(void);\n");

    /* walk through all active services */
    while ((tmp = active_db_next(tmp)))
    {
        D_("try_start_wait_deps(): re-trying with %s.\n", tmp->name);

        /* Look at status */
        switch (tmp->a_status)
        {
            case WAITING_FOR_STOP_DEP:
            case STOP_DEP_MET:
                /* if succeed to stop some, remember this */
                try_to_stop_service(tmp);
                break;
            case WAITING_FOR_START_DEP:
            case START_DEP_MET:
                /* if succeed to start some, remember this too */
                try_to_start_service(tmp);
                break;
            default:
                break;
        }
    }

}

/* enter a new runlevel */
int stop_all(void)
{
    active_h *ser = NULL;

    D_("stop_all();\n");
    while ((ser = active_db_next(ser)))
    {
        if (ser)
            stop_service(ser);
    }
    return (TRUE);
}

/* return TRUE if service is okay to load */
static int service_blacklisted(char *name)
{
    int i;

    assert(name);
    /* walk through arguments looking for this dep to be blacklisted */
    for (i = 0; (g.Argv)[i]; ++i)
    {
        /* if we got a match */
        if ((g.Argv)[i][0] == '-' && !strcmp((g.Argv)[i] + 1, name))
        {
            return (TRUE);
        }
    }
    return (FALSE);
}

/* start new process named */
int start_new_service_named(char *service)
{
    active_h *to_start = NULL;

    assert(service);
    load_to_active(&to_start, service);

    if (to_start)
        return (start_service(to_start));

    return (FALSE);
}

/* Load dependencies */
static int load_all_deps(active_h * service_starting)
{
    char *dep = NULL;
    int deps = 0;

    assert(service_starting);
    D_("load_all_deps(%s);\n", service_starting->name);

    /* if there are no dependencies dont do anything */
    if (!(deps = active_db_count_type(STRING_NEED, service_starting)))
    {
        D_("load_all_deps(%s): service has no dependencies\n",
           service_starting->name);
        service_starting->depends = NULL;
        return (TRUE);
    }

    D_("allocating for %i strings on %s.\n", deps, service_starting->name);
    service_starting->depends = calloc(deps + 1, sizeof(char *));
    deps = 0;
    while ((dep =
            active_db_get_next_string(STRING_NEED, service_starting, dep)))
    {
        D_("a dep is:\"%s\" #%i.\n", dep, deps);
        load_to_active(&service_starting->depends[deps], strdup(dep));
        deps++;
    }
    service_starting->depends[deps] = NULL;
    D_("load_all_deps(%s): Returning success.\n", service_starting->name);
    return (TRUE);

}

/* is all in STRING_USE up? */
static int is_use_up(active_h * service_starting)
{
    active_h *t_a;
    char *string = NULL;

    assert(service_starting);
    D_("is_use_up(%s);\n", service_starting->name);

    while ((string =
            active_db_get_next_string(STRING_USE, service_starting, string)))
    {
        if ((t_a = active_db_find_by_name(string)))
        {
            D_("found %s, :", string);
            switch (t_a->a_status)
            {
                case STARTING:
                case LOADING:
                case WAITING_FOR_START_DEP:
                case START_DEP_MET:
                    D_(" waiting on it... \n");
                    return (FALSE);
                default:
                    D_(" is online ...\n");
                    break;
            }
        }
#ifdef DEBUG
        else
            D_("have use %s, but no souch service found.\n", string);
#endif
    }

    return (TRUE);
}

/* called when a process got killed, identify it, and make a call with a pointer to the process */
void handle_killed_by_pid(pid_t kpid, int r_code)
{
    /* The process that got killed */
    active_h *killd_p = NULL;

    D_("handle_killed_by_pid(%i);\n", kpid);

    /* don't do anyting on this pid */
    if (kpid <= 1)
        return;

    /* Look in process database for a match */
    if (!(killd_p = active_db_find_by_pid(kpid)))
    {
        D_("handle_killed_by_pid(%i): No match!\n", kpid);
        return;
    }

    D_("handle_killed_by_pid(%i): found service \"%s\"...\n", kpid,
       killd_p->name);

    if (killd_p->start_process && killd_p->start_process->pid == kpid)
    {
        D_("handle_killed_by_pid(%i): seems to be a start process killed...\n", kpid);
        killd_p->start_process->r_code = r_code;
        handle_killed_start(killd_p);
        return;
    }
    else if (killd_p->stop_process && killd_p->stop_process->pid == kpid)
    {
        D_("handle_killed_by_pid(%i): seems to be a stop process killed...\n",
           kpid);
        killd_p->stop_process->r_code = r_code;
        handle_killed_stop(killd_p);
        return;
    }

    F_("handle_killed_by_pid(%i): traced killd pid to a serivce but can't finde servicetype!\n", kpid);
    return;
}

/* to do when a start action (or a deamon) die */
static void handle_killed_start(active_h * killed_service)
{
    assert(killed_service);
    assert(killed_service->name);

    D_("handle_killed_start(%s);\n", killed_service->name);

    /* print a error message if r_code is positive */
    if (killed_service->start_process->r_code)
        out_service_error(killed_service->name,
                          killed_service->start_process->r_code);

    /* if this service has the request to stop, this is good */
    if (killed_service->a_status == STOPPING)
    {
        mark_service(killed_service, STOPPED);
        active_db_free(killed_service);
        return;
    }

    /* load service if have been lost */
    if (!get_service(killed_service))
        return;

    /* If this service is a service that is marked up when start
     * process exits, detect this */
    if (active_db_is(ENABLED_UP_WHEN_DONE, killed_service) ||
        ((active_db_is(STRING_START, killed_service) ||
          active_db_is(STRING_START_SCRIPT, killed_service)) &&
         !active_db_is(ENABLED_UP_WHEN_STARTED, killed_service)))
    {

        /* if return code, is positive */
        if (killed_service->start_process->r_code)
            mark_service(killed_service, FAIL_STARTING);
        else
            mark_service(killed_service, DONE);

        process_db_free(&killed_service->start_process);
        return;
    }

    /* if this service is set to respawn, do it. */
    if (active_db_is(ENABLED_RESPAWN, killed_service))
    {
        mark_service(killed_service, RESPAWNING);
        process_db_free(&killed_service->start_process);
        /* respawn */
        start_service(killed_service);
        return;
    }

    /* else, mark as stop, and forgot */
    mark_service(killed_service, STOPPED);
    active_db_free(killed_service);

}

static void mark_service(active_h * service, e_a_status status)
{
    assert(service);
    /* if already set */
    if (service->a_status == status)
    {
        D_("warning, this status %i is already set on %s!\n", status,
           service->name);
        return;
    }

    /* make an interrupt */
    g.interupt = TRUE;

    D_("mark_service(%s, %s);\n", service->name,
       active_db_get_status_string(status));

    /* set status and time */
    service->a_status = status;
    service->time_got_status = time(0);
    history_add_values(service->name, service->time_got_status, status);

    /* call all dynamicly modules that want to know about this */
    initng_load_module_state_changed(service);

    print_output(service);

    /* check if this is the final service */
    if (strcmp(service->name, g.runlevel) == 0 && service->a_status == DONE)
        system_up();
}

/* to do when a stop action die */
static void handle_killed_stop(active_h * killed_service)
{
    assert(killed_service);

    D_("handle_killed_stop(%s);\n", killed_service->name);

    /* free stopped process data */
    process_db_free(&killed_service->stop_process);

    /* if there is no start process, count this as the end of service */
    if (!killed_service->start_process)
    {
        /* mark service stopped */
        mark_service(killed_service, STOPPED);

        /* free service, if stopped = disappear from active list */
        active_db_free(killed_service);
    }
}

/* this function will send term signal to a service if it was marked 30 seconds or more */
void term_stopping_services(void)
{
    active_h *to_term = NULL;

    while ((to_term = active_db_next(to_term)))

        if (to_term &&
            to_term->a_status == STOPPING &&
            to_term->time_got_status &&
            to_term->start_process && to_term->start_process->pid)
        {
            /*printf("tome_got_status=%i, time now=%i\n",(int) to_term->time_got_status, (int)time(0)); */

            if (to_term->time_got_status + 9 < time(0))
            {
                printf("Warning, terminating %s on pid %i!\n", to_term->name,
                       to_term->start_process->pid);
                kill(-to_term->start_process->pid, SIGKILL);
            }
            else
            {
                alarm(10);                  /* if there are stopping processes, schedule a new alarm */
            }
        }

}

/* S T O P     S E R V I C E */
e_a_status stop_service(active_h * service_to_stop)
{
    active_h *current = NULL;

    assert(service_to_stop);

    D_("stop_service(%s);\n", service_to_stop->name);

    /* check if service status, is really up */
    if (service_to_stop->a_status != DONE &&
        service_to_stop->a_status != RUNNING)
        return (FAIL_STOPPING);

    /* set stopping */
    mark_service(service_to_stop, STOP_MARKED);

    /* mark all services depending on this service stop */
    while ((current = active_db_next_dep(service_to_stop, current)))
    {
        D_("%s have to stop %s.\n", service_to_stop->name, current->name);
        stop_service(current);
    }

    /* if there was a service depending on this, mark this service, Waiting for stop dep. */
    if (current)
        mark_service(service_to_stop, WAITING_FOR_STOP_DEP);
    else
        mark_service(service_to_stop, STOP_DEP_MET);

    return (service_to_stop->a_status);
}

/* do stop service */
static int try_to_stop_service(active_h * service_to_stop)
{
    active_h *current = NULL;

    assert(service_to_stop);

    D_("try_to_stop_service(%s): status :\"%s\"\n",
       service_to_stop->name,
       active_db_get_status_string(service_to_stop->a_status));

    /* if there are services depending on this one still running, return false and still try */
    while ((current = active_db_next_dep(service_to_stop, current)))
    {
        if (current->a_status == DONE ||
            current->a_status == RUNNING ||
            current->a_status == STOP_MARKED ||
            current->a_status == STOPPING ||
            current->a_status == WAITING_FOR_STOP_DEP ||
            current->a_status == STOP_DEP_MET)
            return (FALSE);                 /* wait */
    }

    /* check if it has a stop process */
    if (service_to_stop->stop_process)
    {
        F_("try_to_stop_service(%s): Function already have a stop process! Aborting!\n", service_to_stop->name);
        mark_service(service_to_stop, FAIL_STOPPING);
        return (FALSE);
    }

    /* F I N A L Y   S T O P */
    mark_service(service_to_stop, STOPPING);

    /* check that service data is in memory */
    if (!get_service(service_to_stop))
    {
        mark_service(service_to_stop, FAIL_STOPPING);
        return (FALSE);
    }

    /* T R Y 1 , USING THE stop script/program */
    if (active_db_is(STRING_STOP, service_to_stop))
    {
        if (!stop_execute(service_to_stop))
        {
            F_("try_to_stop_service(%s): FAIL EXEUTING initng_execute_fork_and_exec!\n", service_to_stop->name);
            mark_service(service_to_stop, FAIL_STOPPING);
            return (FALSE);
        }

        if (service_to_stop->start_process
            && service_to_stop->start_process->pid)
            alarm(10);                      /* the alarm will try to kill stopping services */
        return (TRUE);
    }

    if (active_db_is(STRING_STOP_SCRIPT, service_to_stop))
    {
        if (!stop_script_execute(service_to_stop))
        {
            F_("try_to_stop_service(%s): FAIL EXEUTING initng_execute_fork_and_exec!\n", service_to_stop->name);
            mark_service(service_to_stop, FAIL_STOPPING);
            return (FALSE);
        }

        if (service_to_stop->start_process
            && service_to_stop->start_process->pid)
            alarm(10);                      /* the alarm will try to kill stopping services */

        return (TRUE);
    }

    /* T R Y 2 , Kill start_process */
    if (service_to_stop->start_process
        && service_to_stop->start_process->pid > 0)
    {
        D_("try_to_stop(%s): REAL KILLING PID %i!\n",
           service_to_stop->name, service_to_stop->start_process->pid);
        kill(-service_to_stop->start_process->pid, SIGTERM);
        alarm(10);                          /* the alarm will try to kill stopping services */
        /*kill(service_to_stop->start_process->pid,SIGKILL); */
        return (TRUE);
    }

    /* if this is a empty dep service whe show up here */

    /* remember this in history */
    mark_service(service_to_stop, STOPPED);

    /* free service, and forget */
    active_db_free(service_to_stop);

    return (TRUE);
}

static void system_up(void)
{
    start_new_service_named("up");
    out_system_up();

}
