
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* malloc() free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <termios.h>
#include <stdio.h>

#include "initng_global.h"
#include "colorprint.h"
#include "main.h"
#include "initng_signal.h"
#include "initng_history.h"
#include "remote_control.h"
#include "initng_handler.h"
#include "initng_execute.h"                 /* new_environ() */
#include "active_db.h"
#include "initng_load_module.h"

extern char **environ;

static void ready_to_quit(void);
static void su_login(void);
static void start_initiating_service(void);

/*
 * Poll if there are any processes left, and if not quit initng
 */
static void ready_to_quit(void)
{
#ifdef REALLY_DEBUG

    /* Track the count of remaining processes */
    D_("ready_to_quit(): %i process(ess) remaining.\n", active_db_count(0));
    {
        active_h *tmp = NULL;

        while ((tmp = active_db_next(tmp)))
        {
            D_("ready_to_quit(): ********  %s status: %i left    *******\n",
               tmp->name, tmp->a_status);
        }
    }
#endif

    /* If the last process has died, quit initng */
    if ((active_db_count(NONE)) < 1)
    {
        out_all_services_exited();          /* Notify of runlevel shutdown */

#ifdef DEBUG
        if (g.runlevel)
        {
            D_("ready_to_quit(): Restarting from service/runlevel: %s\n\n",
               g.runlevel);
        }
#endif

        switch (g.when_out)
        {
            default:
                F_("Error: invalid command (code %d)\n", g.when_out);
                /* FALLTHROUGH */
            case THEN_QUIT:
                exit_initng(0);
                break;
            case THEN_RESTART:
                restart_initng();
                break;
            case THEN_START_DEFAULT:
                /* start first service */
                start_initiating_service();
                break;
            case THEN_SULOGIN:
                su_login();
                break;
            case THEN_NEW_RUNLEVEL:
                if (g.runlevel)
                {
                    /* if runlevel is set to sulogin, do that */
                    if ((strcasecmp(g.runlevel, "sulogin") == 0))
                        su_login();
                    else
                        start_new_service_named(g.runlevel);
                }
                break;
        }
    }
}

/*
 * Exiting on an initialization failure is not a good idea for init
 */
static void su_login(void)
{
    char *argv[] = { "sulogin", NULL };
    execve("/sbin/sulogin", argv, new_environ(NULL));   /* launch sulogin */
    _exit(1);
}

/*
 * Start initial services
 */
static void start_initiating_service(void)
{
    int i;
    int a_count = 0;            /* counts orders from argv to start */

    /* check with argv which service to start initiating */
    for (i = 1; i < g.Argc; i++)
    {

        /* if succeed to load this service */
        if (start_new_service_named(g.Argv[i]))
            a_count++;
    }
}

/*
 * %%%%%%%%%%%%%%%%%%%%   main ()   %%%%%%%%%%%%%%%%%%%%
 */
#ifdef BUSYBOX
int init_main(int argc, char *argv[])
{
#else
int main(int argc, char *argv[])
{
#endif
    time_t last;                /* save the time here */
    int got_remote = FALSE;

    /* initiate global variables */
    i_g();
#ifdef DEBUG
    int loop_counter = 0;       /* a common used for counter */
#endif

    D_("main();\n");

    /* set nice level */

    /* nice(INITNG_NICELEVEL); */

    /* Tell the user he is welcome */
    out_welcome();

    /* get the time */
    last = time(NULL);

    /* Null */
    history_db_init();

    /* Load script parser modules */
    initng_load_module("/lib/initng/parsers/libezxmlparser.so.0.0");
    initng_load_module("/lib/initng/parsers/libiparser.so.0.0");

    /*initng_load_module("/lib/initng/parsers/libxmlparser.so.0.0"); */
    initng_load_module("/lib/initng/parsers/librlparser.so.0.0");

    /* Pointer to starting arguments, this makes them GLOBAL */
    g.Argv = argv;
    g.Argc = argc;

    /* See if we are the first process born in the system */
    if (getpid() == 1)
    {
        int fd;                 /* /dev/console */

        g.i_am_init = TRUE;
        g.when_out = THEN_RESTART;
        reboot(RB_DISABLE_CAD);             /* Disable Ctrl + Alt + Delete */

        /* Try to open the console, but don't control it */
        if ((fd = open("/dev/console", O_RDWR | O_NOCTTY)) > 0)
        {
            D_("Opened /dev/console. Setting terminal options.\n");
            ioctl(fd, KDSIGACCEPT, SIGWINCH);   /* Accept signals from 'kbd' */
            close(fd);                      /* Like Ctrl + Alt + Delete signal? */
        }
        else
        {
            D_("Failed to open /dev/console. Setting options anyway.\n");
            ioctl(0, KDSIGACCEPT, SIGWINCH);    /* Accept signals from 'kbd' */
        }

        /* TODO: this block may be incorrect or incomplete */
        {
            struct termios tty;
            int fd;

            /*
             * TODO: /dev/console may still be open from before. Also, if it
             * fails to open, why try to finish the block?
             */
            if ((fd = open("/dev/console", O_RDWR | O_NOCTTY)) < 0)
            {
                F_("main(): can't open /dev/console.\n");
            }

            (void) tcgetattr(fd, &tty);

            tty.c_cflag &= CBAUD | CBAUDEX | CSIZE | CSTOPB | PARENB | PARODD;
            tty.c_cflag |= HUPCL | CLOCAL | CREAD;

            tty.c_cc[VINTR] = 3;            /* ctrl('c') */
            tty.c_cc[VQUIT] = 28;           /* ctrl('\\') */
            tty.c_cc[VERASE] = 127;
            tty.c_cc[VKILL] = 24;           /* ctrl('x') */
            tty.c_cc[VEOF] = 4;             /* ctrl('d') */
            tty.c_cc[VTIME] = 0;
            tty.c_cc[VMIN] = 1;
            tty.c_cc[VSTART] = 17;          /* ctrl('q') */
            tty.c_cc[VSTOP] = 19;           /* ctrl('s') */
            tty.c_cc[VSUSP] = 26;           /* ctrl('z') */

            /*
             *      Set pre and post processing
             */
            tty.c_iflag = IGNPAR | ICRNL | IXON | IXANY;
            tty.c_oflag = OPOST | ONLCR;
            tty.c_lflag = ISIG | ICANON | ECHO | ECHOCTL | ECHOPRT | ECHOKE;

            /*
             *      Now set the terminal line.
             *      We don't care about non-transmitted output data
             *      and non-read input data.
             */
            (void) tcsetattr(fd, TCSANOW, &tty);
            (void) tcflush(fd, TCIOFLUSH);
            (void) close(fd);
        }

    }

    /* Configure signal handlers */
    (void) enable_signals();

    /* Enable control through the 'in' and 'out' FIFO's */
    got_remote = remote_control_open(g.i_am_init);

    /* If no starting services found, use default */
    if (!g.runlevel)
    {
        if (g.i_am_init)
        {
            g.runlevel = strdup("default");
        }
        else
        {
            g.runlevel = strdup("fake-default");
        }
    }

    start_new_service_named(g.runlevel);
    start_initiating_service();             /* Start the first service */

    /* %%%%%%%%%%%%%%%   MAIN LOOP   %%%%%%%%%%%%%%% */
    for (;;)
    {
        time_t now;             /* Store the start time of each loop */

        D_("main():mainloop: %i\n", loop_counter++);

        /* Update current time */
        now = time(NULL);

        /* Check some global variables, set by sigactions */

        /* If we got an interrupt load ctrlaltdel */
        if (initng_signal_got_ctrl_alt_del)
        {
            initng_signal_got_ctrl_alt_del = 0;
            start_new_service_named("ctrlaltdel");
        }

        /* If we got a keybord request load kbreq */
        if (initng_signal_got_keyboard_request)
        {
            initng_signal_got_keyboard_request = 0;
            start_new_service_named("kbreq");
        }

        /*
         * If now is before last, the time was set backwards.
         * If so, compensate the change in the whole process_db
         */
        if (now < last)
        {
            D_("main():mainloop: The clock was reset. Compensating...\n");
            active_db_compensate_time(last - now);
            history_db_compensate_time(last - now);
        }

        /*
         * If we moved more than an hour ahead, work backwards in one hour
         * chunks until we are within one hour of the original.
         *
         * TODO: why is this important?
         */
        while (now - last >= 3600)
        {
            active_db_compensate_time(3600);
            history_db_compensate_time(3600);
            last = last + 3600;
        }
        last = now;                         /* Remember last time */

        /* 
         * Start launching dependencies, until no more services are marked.
         * The dependencies db should continue to stabilize through each run.
         */
        while (g.interupt)
        {
            g.interupt = FALSE;
            try_start_wait_deps();
            D_("main():miniloop (will continue as long as interrupt is set)\n");
        }
        D_("main(): Finished loading dependencies!\n");

        /* Look for any processes that won't DIE */

        /* term_stopping_services(); */

        /*
         * If the remote connection failed before, try again
         */
        if (!got_remote)
            got_remote = remote_control_open(g.i_am_init);

        /*
         * Check if there are any running processes left, otherwise quit
         *
         * TODO: it isn't clear that ready_to_quit() is what stops initng.
         * "ready_to_quit" sounds like a condition test, we should respond to
         * it here with a handler.
         *
         * if( (status = ready_to_quit()) )
         * {
         *     status_change_handler(status);
         * }
         */
        ready_to_quit();

        /*
         * See if anything is happening on the control channel
         *  - wait for 60seconds
         */
        if (!g.interupt)
            remote_control_poll(60000);
    }                                       /* End main loop */
}

void restart_initng(void)
{
    printf("\n\n\n          R E S T A R T I N G,  (Really hot reboot)\n\n");
    execve("/sbin/initng", g.Argv, environ);
}

void exit_initng(int i)
{
    D_("exit_initng();\n");

    initng_load_module_close_all();         /* close all modules */
    history_free_all();                     /* flush history log db */
    remote_control_close();                 /* close control fifos */
    active_db_free_all();                   /* clean process_db */
    service_db_free_all();                  /* clean service db */

    if (g.runlevel)
        free(g.runlevel);

    /* If i == 99, we only cleanup everything, don't exit */
    if (i != 99)
        exit(i);
}
